/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.analysis;

import jalview.api.analysis.ScoreModelI;
import jalview.api.analysis.SimilarityParamsI;
import jalview.bin.Console;
import jalview.datamodel.AlignmentView;
import jalview.datamodel.Point;
import jalview.datamodel.SequenceI;
import jalview.datamodel.SequenceGroup;
import jalview.gui.PairwiseAlignPanel;
import jalview.gui.PaSiMapPanel;
import jalview.math.Matrix;
import jalview.math.MatrixI;
import jalview.math.MiscMath;
import jalview.viewmodel.AlignmentViewport;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Enumeration;

/**
 * Performs Principal Component Analysis on given sequences
 * @AUTHOR MorellThomas 
 */
public class PaSiMap implements Runnable
{
  /*
   * inputs
   */
  final private AlignmentViewport seqs;

  final private ScoreModelI scoreModel;

  final private SimilarityParamsI similarityParams;

  final private byte dim = 3;

  final private int openCost = 100;

  final private int extendCost = 5;

  /*
  * other
  */
  private long total;	// total number of combinations

  private long count;	// current iteration number

  /*
   * outputs
   */
  private PairwiseAlignPanel alignment;

  private MatrixI pairwiseScores;

  private MatrixI eigenMatrix;

  /**
   * Constructor given the sequences to compute for, the similarity model to
   * use, and a set of parameters for sequence comparison
   * 
   * @param sequences
   * @param sm
   * @param options
   */
  public PaSiMap(AlignmentViewport sequences, ScoreModelI sm,
          SimilarityParamsI options)
  {
    this.seqs = sequences;
    this.scoreModel = sm;
    this.similarityParams = options;
    int nseqs = seqs.getAlignment().getHeight();
    this.total = (long) MiscMath.permutations(nseqs, 2) / 2;	// n!/(n-r)!
    this.count = 0l;
  }

  /**
   * Returns Eigenvalue
   * 
   * @param i
   *          Index of diagonal within matrix
   * 
   * @return Returns value of diagonal from matrix
   */
  public double getEigenvalue(int i)
  {
    return eigenMatrix.getD()[i];
  }

  /**
   * Returns coordinates for each datapoint
   * 
   * @param l
   *          DOCUMENT ME!
   * @param n
   *          DOCUMENT ME!
   * @param mm
   *          DOCUMENT ME!
   * @param factor ~ is 1
   * 
   * @return DOCUMENT ME!
   */
  public Point[] getComponents(int l, int n, int mm, float factor)
  {
    Point[] out = new Point[getHeight()];

    for (int i = 0; i < out.length; i++)
    {
      float x = (float) component(i, l) * factor;
      float y = (float) component(i, n) * factor;
      float z = (float) component(i, mm) * factor;
      out[i] = new Point(x, y, z);
    }

    return out;
  }

  /**
   * DOCUMENT ME!
   * 
   * @param n
   *          DOCUMENT ME!
   * 
   * @return DOCUMENT ME!
   */
  public double[] component(int n)
  {
    // n = index of eigenvector
    double[] out = new double[getWidth()];

    for (int i = 0; i < out.length; i++)
    {
      out[i] = component(i, n);
    }

    return out;
  }

  /**
   * DOCUMENT ME!
   * 
   * @param row
   *          DOCUMENT ME!
   * @param n
   *          DOCUMENT ME!
   * 
   * @return DOCUMENT ME!
   */
  double component(int row, int n)
  {
    return eigenMatrix.getValue(row, n);
  }

  /**
   * Answers a formatted text report of the PaSiMap calculation results (matrices
   * and eigenvalues) suitable for display
   * 
   * @return
   */
  public String getDetails()
  {
    StringBuilder sb = new StringBuilder(1024);
    sb.append("PaSiMap calculation using ").append(scoreModel.getName())
            .append(" sequence similarity matrix\n========\n\n");
    PrintStream ps = wrapOutputBuffer(sb);

    /*
     * coordinates matrix, with D vector
     */
    sb.append(" --- Pairwise correlation coefficients ---\n");
    pairwiseScores.print(ps, "%8.6f ");
    ps.println();

    sb.append(" --- Eigenvalues ---\n");
    eigenMatrix.printD(ps, "%15.4e");
    ps.println();

    sb.append(" --- Coordinates ---\n");
    eigenMatrix.print(ps, "%8.6f ");
    ps.println();

    return sb.toString();
  }

  /**
   * Performs the PaSiMap calculation
   *
   * creates a new gui/PairwiseAlignPanel with the input sequences (AlignmentViewport)
   * uses analysis/AlignSeq to creatue the pairwise alignments and calculate the AlignmentScores (float for each pair)
   * gets all float[][] scores from the gui/PairwiseAlignPanel
   * checks the connections for each sequence with AlignmentViewport seqs.calculateConnectivity(float[][] scores, int dim) (from analysis/Connectivity) -- throws an Exception if insufficient
   * creates a math/MatrixI pairwiseScores of the float[][] scores
   * copys the scores and fills the diagonal to create a symmetric matrix using math/Matrix.fillDiagonal()
   * performs the analysis/ccAnalysis with the symmetric matrix
   * gets the eigenmatrix and the eigenvalues using math/Matrix.tqli()
   */
  @Override
  public void run()
  {
    try
    {
for (SequenceI see : seqs.getAlignment().getSequencesArray())
{
System.out.println(see.getName());
}

/*
      int nSeqs = seqs.getAlignment().getHeight();
      float[][] scores = new float[nSeqs][nSeqs];	// rows, cols

      int nSplits = 1;
      while (((float) nSeqs / nSplits) > 20f)		// heap full at 341
      //while (((float) nSeqs / nSplits) > 5f)		// heap full at 341
	nSplits++;
      int splitSeqs = (int) Math.ceil((float) nSeqs / nSplits);
System.out.println(String.format("%d -> %d splits into %d seqs", nSeqs, nSplits, splitSeqs));

      int[] splitIndices = new int[nSplits];
      for (int i = 0; i < nSplits; i++)
      {
	splitIndices[i] = splitSeqs * (i + 1); 	//exclusive!!
      }

      HashMap<int[], Float> valuesForScores = splitCombineAndAlign(seqs.getAlignment().getSequencesArray(), splitIndices, splitSeqs);

      for (int[] coords : valuesForScores.keySet())
      {
	scores[coords[0]][coords[1]] = valuesForScores.get(coords);
      }
pairwiseScores = new Matrix(scores);
pairwiseScores.print(System.out, "%1.4f ");
*/

      alignment = new PairwiseAlignPanel(seqs, true, 100, 5);
      float[][] scores = alignment.getAlignmentScores();	//bigger index first -- eg scores[14][13]

      //Hashtable<SequenceI, Integer> connectivity = seqs.calculateConnectivity(scores, dim);

      pairwiseScores = new Matrix(scores);
pairwiseScores.print(System.out, "%1.4f ");
      pairwiseScores.fillDiagonal();

      eigenMatrix = pairwiseScores.copy();

      ccAnalysis cc = new ccAnalysis(pairwiseScores, dim);
      eigenMatrix = cc.run();

    } catch (Exception q)
    {
      Console.error("Error computing PaSiMap:  " + q.getMessage());
      q.printStackTrace();
    }
  }

  /**
  * aligns sequences in splits
  * Splits each split into halves and aligns them agains halves of other splits
  *
  * @param seqs
  * @param i ~ indices of split
  * @param s ~ sequences per split
  *
  * @return  a map of where to put in scores, value ~ scores[n][m] = v
  **/
  protected HashMap<int[], Float> splitCombineAndAlign(SequenceI[] seqArray, int[] i, int s)
  {
    HashMap<int[], Float> result = new HashMap<int[], Float>();

    int[][] allGroups = new int[i.length][s];
    for (int g = 0; g < i.length; g++)	// group g
    {
      int e = 0;	// index going through allGroups[][e]
      for (int j = g * s; j < i[g]; j++)  // goes through all numbers in one group
      {
        allGroups[g][e++] = j >= seqArray.length ? -1 : j;
      }
    }
    
    int g = 0;	// group count
    for (int[] group : allGroups)
    {
      ArrayList<SequenceI> sg = new ArrayList<SequenceI>();
      //SequenceGroup sg = new SequenceGroup();
      for (int index : group)
      {
	if (index == -1)
	  continue;
	//sg.addSequence(seqArray[index], false);
	sg.add(seqArray[index]);
      }
      SequenceI[] sgArray = new SequenceI[sg.size()];
      int k = 0;
      for (SequenceI seq : sg)
      {
	sgArray[k++] = seq;
      }
      //seqs.setSelectionGroup(sg);
      //PairwiseAlignPanel pap = new PairwiseAlignPanel(seqs, true, 100, 5);
      //float[][] scores = pap.getAlignmentScores();	//bigger index first -- eg scores[14][13]
      float[][] scores = simulateAlignment(sgArray);
      for (int s1 = 0; s1 < scores.length; s1++)	// row
      {
	result.put(new int[]{s1 + g * s, s1 + g * s}, Float.NaN);	// self score = Float.NaN
	for (int s2 = 0; s2 < s1; s2++)			// col
	{
System.out.println(String.format("%s x %s -> [%d][%d] (%f)", sgArray[s1].getName(), sgArray[s2].getName(), s1+g*s, s2+g*s, scores[s1][s2]));
	  result.put(new int[]{s1 + g * s, s2 + g * s}, scores[s1][s2]);
	}
      }   
      g++;
    }

    int smallS = (int) Math.ceil((float) s/2);
    int[][] newGroups = new int[i.length * 2][smallS];
    
    g = 0;
    for (int[] group : allGroups)
    {
      int[] split1 = new int[smallS];
      int[] split2 = new int[smallS];
      for (int k = 0; k < group.length; k++)	
      {
	if (k < smallS)
	  split1[k] = group[k];
	else
	  split2[k - smallS] = group[k];
      }
      newGroups[g++] = split1;
      newGroups[g++] = split2;
    }

    // align each subsplit with subsplits from other split groups
    for (int subsplitN = 0; subsplitN < newGroups.length; subsplitN++)
    {
System.gc();
      int c = 1;		// current subsplit block
      while (newGroups[subsplitN][0] > smallS * c)
      {
	c++;
      }
      for (int nextSplit = subsplitN + 1; nextSplit < newGroups.length; nextSplit++)
      {
System.out.println(String.format("subsplitN %d, nextSplit %d, %d > %d * %d", subsplitN, nextSplit, allGroups[(int) Math.floor((float) nextSplit / 2)][0], smallS, c));
	if (allGroups[(int) Math.floor((float) nextSplit / 2)][0] > smallS * c)
	{
          ArrayList<SequenceI> sg = new ArrayList<SequenceI>();
          //SequenceGroup sg = new SequenceGroup();
          for (int index : newGroups[subsplitN])
          {
	    if (index == -1)
	      continue;
	    //sg.addSequence(seqArray[index], false);
	    sg.add(seqArray[index]);
          }
          for (int index : newGroups[nextSplit])
          {
	    if (index == -1)
	      continue;
	    //sg.addSequence(seqArray[index], false);
	    sg.add(seqArray[index]);
          }
          SequenceI[] sgArray = new SequenceI[sg.size()];
          int k = 0;
          for (SequenceI seq : sg)
          {
	    sgArray[k++] = seq;
          }
          //seqs.setSelectionGroup(sg);
          //PairwiseAlignPanel pap = new PairwiseAlignPanel(seqs, true, 100, 5);
          //float[][] scores = pap.getAlignmentScores();	//bigger index first -- eg scores[14][13]
          float[][] scores = simulateAlignment(sgArray);
System.gc();
          for (int s1 = 0; s1 < scores.length; s1++)	// row
          {
	    for (int s2 = 0; s2 < s1; s2++)			// col
	    {
	      if (s1 >= smallS && s2 < smallS)
System.out.println(String.format("%s x %s -> [%d][%d] (%f)", sgArray[s1].getName(), sgArray[s2].getName(), s1+(nextSplit-1)*smallS, s2+subsplitN*smallS, scores[s1][s2]));
	        result.put(new int[]{s1 + (nextSplit-1) * smallS, s2 + subsplitN * smallS}, scores[s1][s2]);
	    }
          }   
	}
      }
    }

    return result;
  }

  /**
  * simulate the alignment of a PairwiseAlignPanel
  *
  * @param seqs
  * @return alignment scores
  */
  protected float[][] simulateAlignment(SequenceI[] seqs)
  {
    float[][] result = new float[seqs.length][seqs.length];
    for (int i = 1; i < seqs.length; i++)
    {
      for (int j = 0; j < i; j++)
      {
	String[] seqStrings = new String[2];
	seqStrings[0] = seqs[i].getSequenceAsString();
	seqStrings[1] = seqs[j].getSequenceAsString();

	AlignSeq as = new AlignSeq(seqs[i], seqStrings[0], seqs[j], seqStrings[1], AlignSeq.PEP, openCost, extendCost);
	as.seqInit(seqs[i], seqStrings[0], seqs[j], seqStrings[1], AlignSeq.PEP);
	as.calcScoreMatrix();
	as.traceAlignmentWithEndGaps();
	as.scoreAlignment();
	//as.printAlignment(System.out);
	result[i][j] = as.getAlignmentScore();
	//as.clear();

System.out.println(String.format("%d / %d (%3.2f)", count, total, (float) ((count / total) * 100)));
	count++;
      }
    }
    return result;
  }

  /**
   * Returns a PrintStream that wraps (appends its output to) the given
   * StringBuilder
   * 
   * @param sb
   * @return
   */
  protected PrintStream wrapOutputBuffer(StringBuilder sb)
  {
    PrintStream ps = new PrintStream(System.out)
    {
      @Override
      public void print(String x)
      {
        sb.append(x);
      }

      @Override
      public void println()
      {
        sb.append("\n");
      }
    };
    return ps;
  }

  /**
   * Answers the N dimensions of the NxM PaSiMap matrix. This is the number of
   * sequences involved in the pairwise score calculation.
   * 
   * @return
   */
  public int getHeight()
  {
    // TODO can any of seqs[] be null?
    return eigenMatrix.height();// seqs.getSequences().length;
  }

  /**
   * Answers the M dimensions of the NxM PaSiMap matrix. This is the number of
   * sequences involved in the pairwise score calculation.
   * 
   * @return
   */
  public int getWidth()
  {
    // TODO can any of seqs[] be null?
    return eigenMatrix.width();// seqs.getSequences().length;
  }

  /**
   * Answers the sequence pairwise similarity scores which were the first step
   * of the PaSiMap calculation
   * 
   * @return
   */
  public MatrixI getPairwiseScores()
  {
    return pairwiseScores;
  }

  public void setPairwiseScores(MatrixI m)
  {
    pairwiseScores = m;
  }

  public MatrixI getEigenmatrix()
  {
    return eigenMatrix;
  }

  public void setEigenmatrix(MatrixI m)
  {
    eigenMatrix = m;
  }

  public PairwiseAlignPanel getAlignments()
  {
    return alignment;
  }

  public String getAlignmentOutput()
  {
    return alignment.getAlignmentOutput();
  }

  public byte getDim()
  {
    return dim;
  }
}
